package nu.helmers.commons.guice;

import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import com.google.inject.AbstractModule;
import com.google.inject.Module;
import com.google.inject.name.Names;

/**
 * Create @Named bindings for all the properties in the given properties file. If you have loaded this module for a
 * property file containing "my.property=myValue", you can constructor-inject the value of this property into an
 * arbitrary (Guice-managed) object as follows:
 * <p/>
 * <pre>
 * @Inject
 * public MyClass(@Named("my.property") String myPropertyValue) { ... // use myPropertyValue }
 * </pre>
 * <p/>
 * See Chapter 3, Page 42, Listing 3-18 of "Google Guice: A Lightweight Dependency Injection Framework".
 */
public class PropertiesModule extends AbstractModule implements Module {

    private final String propertiesFile;

    public PropertiesModule(String propertiesFile) {
        super();
        this.propertiesFile = propertiesFile;
    }

    protected void configure() {
        try {
            Properties properties = loadProperties(this.propertiesFile);
            Names.bindProperties(binder(), properties);
        } catch (Exception e) {
            addError("Couldn't configure properties from " + this.propertiesFile, e);
        }
    }

    private static Properties loadProperties(String name) throws IOException {
        Properties properties = new Properties();
        InputStream is = new Object() {
        }.getClass().getEnclosingClass().getResourceAsStream(name);
        try {
            properties.load(is);
        } catch (IOException e) {
            throw new RuntimeException(e);
        } finally {
            if (is != null) {
                is.close();
            }
        }
        return properties;
    }
}
