package nu.helmers.jbpmtest.wiring.modules;

import nu.helmers.commons.guice.PropertiesModule;
import nu.helmers.jbpmtest.dao.impl.issue.IssueDaoImpl;
import nu.helmers.jbpmtest.dao.issue.IIssueDao;
import nu.helmers.jbpmtest.wiring.providers.HibernateConfigurationProvider;

import org.hibernate.cfg.Configuration;

import com.google.inject.AbstractModule;
import com.google.inject.Module;
import com.google.inject.matcher.Matcher;

/**
 * This module binds all the Dao interfaces to implementations, and makes sure that all methods that are annotated with @Transactional
 * in your client code are wrapped in a transaction. (Note that the dao methods themselves are not transactional; you
 * will normally make the service methods @Transactional.)
 * <p/>
 * To use this jar in your client code, load this {@link DataLayerModule} and load the required properties (as described
 * in jbpmtest-dao.properties) using a {@link PropertiesModule}.
 * <p/>
 * Notice that the properties files inside this jar are only meant for testing purposes. In your web application, the
 * properties are typically defined in the web layer or somewhere outside the war at deployment time!
 */
public class DataLayerModule extends AbstractModule implements Module {

    @SuppressWarnings("unchecked")
    private Matcher<Class> classMatchers;

    private Matcher<Object> methodMatchers;

    /** Default constructor, which creates a transaction for every method annotated with @Transactional. */
    public DataLayerModule() {
        super();
    }

    /**
     * Use this constructor to create a data layer with a custom {@link PersistenceModule}.
     * 
     * @see PersistenceModule
     */
    @SuppressWarnings("unchecked")
    public DataLayerModule(Matcher<Class> classes, Matcher<Object> methods) {
        super();
        if (classes == null || methods == null) {
            throw new IllegalArgumentException("All arguments of this constructor are required.");
        }
        this.classMatchers = classes;
        this.methodMatchers = methods;
    }

    @Override
    protected void configure() {
        bind(Configuration.class).toProvider(HibernateConfigurationProvider.class);

        /*
         * This makes the hibernate Session and SessionFactory available to Guice. See
         * http://www.wideplay.com/webextensions%3A%3Ajpaintegration2 and
         * com.wideplay.warp.persist.hibernate.HibernatePersistenceStrategy.
         */
        if (this.classMatchers != null) {
            // Used if methods without @Transactional should be wrapped in a transaction.
            install(new PersistenceModule(this.classMatchers, this.methodMatchers));
        } else {
            install(new PersistenceModule());
        }
        bindDaos();
    }

    private void bindDaos() {
        bind(IIssueDao.class).to(IssueDaoImpl.class);
    }
}
