package nu.helmers.jbpmtest.dao.impl;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertNotSame;
import static junit.framework.Assert.fail;

import java.util.List;

import nu.helmers.jbpmtest.dao.test.db.DatabaseHelper;
import nu.helmers.jbpmtest.domain.DomainObject;
import nu.helmers.jbpmtest.hibernate.AbstractGenericDao;
import nu.helmers.jbpmtest.hibernate.IGenericDao;
import nu.helmers.jbpmtest.wiring.modules.DataLayerIntegrationTestModule;

import org.junit.After;
import org.junit.Before;

import com.google.inject.Guice;
import com.google.inject.Injector;

/**
 * Test the generic methods on the {@link AbstractGenericDao}, for a given domain object type. This is to avoid a lot of
 * copy-paste of test methods for all the different concrete daos. In the concrete dao integration tests, we only have
 * to test the methods that the specific dao adds to the generic dao.
 * <p/>
 * {@link DataLayerIntegrationTestModule} ensures that all dao-methods are wrapped in their own transaction, so two
 * subsequent calls to the dao under test will always hit the database. (dao methods are normally <i>not</i>
 * transactional because they're not annotated with @Transactional).
 */
public abstract class GenericDaoTest<T> {

    private DatabaseHelper dbHelper;

    public DatabaseHelper getDbHelper() {
        return dbHelper;
    }

    /** You can call {@link #testPersistAndGet(IGenericDao, Object)} in your implementation */
    public abstract void testPersistAndGet();

    /** You can call {@link #testGetThrowsExceptionForNonExistingId(IGenericDao)} in your implementation */
    public abstract void testGetThrowsExceptionForNonExistingId();

    /** You can call {@link #testFindAll(IGenericDao, Object...)} in your implementation */
    public abstract void testFindAll();

    /** You can call {@link #testDelete(IGenericDao, Object)} in your implementation */
    public abstract void testDelete();

    /** You can call {@link #testCount(IGenericDao, Object...)} in your implementation */
    public abstract void testCount();

    /** This base class does not offer a generic way to test updates. Write it yourself. */
    public abstract void testUpdate();

    @Before
    public void setup() {
        Injector injector = Guice.createInjector(new DataLayerIntegrationTestModule());
        injector.injectMembers(this);
        dbHelper = injector.getInstance(DatabaseHelper.class);
    }

    @After
    public void teardown() {
        try {
            dbHelper.truncate();
        } finally {
            dbHelper = null;
        }
    }

    protected void testPersistAndGet(IGenericDao<T> dao, T newEntity) {
        dao.persist(newEntity);
        Long id = ((DomainObject) newEntity).getId();

        T persistedEntity = dao.get(id);
        assertNotSame(newEntity, persistedEntity); // Make sure it's not the original object (from cache).
        assertEquals(newEntity, persistedEntity);
    }

    protected void testGetThrowsExceptionForNonExistingId(IGenericDao<T> dao) {
        // Search for a big Id, to make sure that the Id is not used by inserted reference data.
        try {
            dao.get(99999L);
        } catch (RuntimeException r) {
            return;
        }
        fail("Expected RuntimeException not thrown.");
    }

    protected void testFindAll(IGenericDao<T> dao, T... newEntities) {
        for (T entity : newEntities) {
            dao.persist(entity);
        }
        List<T> list = dao.findAll();
        assertEquals(newEntities.length, list.size());
    }

    protected void testDelete(IGenericDao<T> dao, T newEntity) {
        // insert one entity.
        dao.persist(newEntity);
        List<T> list = dao.findAll();
        assertEquals(1, list.size());

        // and delete it again.
        dao.delete(list.get(0));
        list = dao.findAll();
        assertEquals(0, list.size());
    }

    protected void testCount(IGenericDao<T> dao, T... newEntities) {
        for (T entity : newEntities) {
            dao.persist(entity);
        }
        assertEquals(newEntities.length, dao.count());
    }

}
