package nu.helmers.jbpmtest.logging;

import java.util.ArrayList;

import nu.helmers.jbpmtest.domain.DomainObject;
import nu.helmers.jbpmtest.service.exception.ServiceException;

import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;

/**
 * Interceptor that logs the entrance and exit of methods, and logs exceptions thrown by the method.
 */
public class LogInterceptor implements MethodInterceptor {

    private final Logger logger;

    public LogInterceptor(Logger logger) {
        this.logger = logger;
    }

    public Object invoke(MethodInvocation invocation) throws Throwable {
        String methodName = invocation.getMethod().getName();
        try {
            logMethodEntry(invocation);
            logParameters(invocation);
            Object result = invocation.proceed();
            logMethodExit(methodName);
            logReturnValue(result);
            return result;
        } catch (ServiceException e) {
            logger.info("END {} with service exception (this can happen during normal operation)", methodName);
            throw e;
        } catch (Throwable e) {
            logger.error("END {} (with exception!)", methodName);
            // no {}: this method has a message, not a format string 
            logger.error("Throwing:", e);
            throw e;
        }
    }

    /** Log method; show "entity.id=value" for BaseEntity objects, and show "?" for all other types. */
    private void logMethodEntry(MethodInvocation invocation) {
        ArrayList<String> params = buildPrintableParameterList(invocation);
        logger.info("BEGIN[{}] {}({})", new Object[] { userName(), invocation.getMethod().getName(),
                StringUtils.join(params, ',') });
    }

    private String userName() {
        final String UNKNOWN = "--";
        return UNKNOWN;
        // TODO implement as soon as we introduce authorisation.
    }

    private ArrayList<String> buildPrintableParameterList(MethodInvocation invocation) {
        ArrayList<String> params = new ArrayList<String>();
        Object[] actuals = invocation.getArguments();
        for (int i = 0; i < actuals.length; i++) {
            addActualToPrintableParameterList(params, actuals[i]);
        }
        return params;
    }

    private void addActualToPrintableParameterList(ArrayList<String> params, Object actual) {
        if (actual instanceof DomainObject) {
            DomainObject domainObject = (DomainObject) actual;
            params.add(domainObject.getClass().getSimpleName() + ".id=" + domainObject.getId());
        } else {
            params.add("?");
        }
    }

    /** Log toString() of each argument (only at DEBUG level) */
    private void logParameters(MethodInvocation invocation) {
        if (logger.isDebugEnabled()) {
            Class<?>[] formals = invocation.getMethod().getParameterTypes();
            Object[] actuals = invocation.getArguments();
            logger.debug("Arguments: ");
            // in case of varargs, formals may be shorter than actuals,
            // therefore we use two loops to properly output the actuals.
            for (int i = 0; i < formals.length; i++) {
                logger.debug("   {} = {}", formals[i], actuals[i]);
            }
            for (int i = formals.length; i < actuals.length; i++) {
                logger.debug("   (vararg) {}", actuals[i]);
            }
        }
    }

    /** Log method exit */
    private void logMethodExit(String methodName) {
        logger.info("END {}", methodName);
    }

    /** Log the return value (only at DEBUG level) */
    private void logReturnValue(Object result) {
        if (logger.isDebugEnabled()) {
            logger.debug("Returning: {}", result);
        }
    }

}
